##################################################
## CTSM-R @: www.ctsm.info
## Simulation of the state
## By Jan Kloppenborg Moller (jkmo@dtu.dk)
## Please read the README.txt file
## License: GPL v2
##################################################
rm(list=ls())

## Load model
load(file="Model6.Rdata")

################################################
## Control
################################################

## set range for cantrol variables 
Range<-list(P=c(0.3,2.5),Q=c(0.4,3.5))

## load ctsm library
library(ctsmr)
# New model
model <- ctsm()

## Set model and parameters 
model<-Model6$model
fit<-Model6$fit

ctsmr:::Compile(model)

## Ctsm setting
model$options$initialVarianceScaling<-exp(-10)

par<-Model6$opt$par   # set parameters 
fit$xm[names(par)]<-par    # Parameters in fit


# Link function to ensure control variable in the range of Range
link2<-function(pars,t,Range){
  # presure and crossflow quadratic (before transformation) in time
    legendre<-function(x,pars){
        x <- x-min(x)
        x<-2*(x-0.5*diff(range(x)))/(diff(range(x)))
        pars[1]+pars[2]*x+pars[3]*0.5*(3*x^2-1)
  }
  Q<-1/(1+exp(-legendre(t,pars[c("q0","q1","q2")])))*
    diff(Range$Q)+Range$Q[1]
  P<-1/(1+exp(-legendre(t,pars[c("p0","p1","p2")])))*
    diff(Range$P)+Range$P[1]
  return(list(P=P,Q=Q))
}


data1<-data.frame(t=seq(0,100,by=5),y=0,DP=0.3,Q=1,VQ2=0,OLQ=0,OLP=0,Rm=0.58)


## Simulated flux and cross flow
sim <- simulate(fit, newdata = data1, first=TRUE)


simInt<-function(pars,Range,t0,tmax,deltat,lambda,fit,init.var,link,z0,I0){
  # Create input and simulate from 0 to T with ctsm, i.e. mean value simulation 
  fit$xm["z0"]<-z0 # Set initial value
  fit$xm["I0"]<-I0 # Set initial value
  t<-seq(t0,tmax,by=deltat) # sequence of times
  N<-length(t) # Nunber of time points

  # control variables
  PQ<-link(pars,t,Range) 
  P<-PQ$P;Q<-PQ$Q


  # dataframe for simulation with ctsm
  data<-data.frame(t=t,y=0,DP=P,Q=Q,VQ2=0,OLQ=0,OLP=0,Rm=0.58)

  # Simulated flux and cross flow
  sim <- simulate(fit, newdata = data, first=TRUE)
  
  # Total integrated flux and standard deviation for integrated flux
  IntJ<- sim$state$sim[N,"I"]/3600
  SdIntJ<-(sim$state$sd[N,"I"])/3600
  # return a list with
  # sim: simulation (mean values and standard deviations)
  # IntJ: total integrated flux
  # data: data used for simulation
  # P,Q: control variables
  # SdIntJ: standard deviation of integrated flux
  return(list(sim=sim,IntJ=IntJ,data=data,P=P,Q=Q,SdIntJ=SdIntJ))
}

##################################################
## Objective function for control
obj.fun<-function(pars,Range,t0,tmax,deltat, IntJ0, lambda,fit,init.var,link,z0,I0){
  # Objective function (integral over crossflow + deviation from target)
  sim<-simInt(pars,Range,t0,tmax,deltat,lambda,fit,init.var,link,z0,I0)
    return(sum((sim$data$Q)^3)/3600*diff(sim$data$t)[1]+
           lambda*(sim$IntJ- qnorm(0.5)*sim$SdIntJ-IntJ0)^2)
}



##################################################
## Example
IntJ0<-3
lambda<-1e5 # penalty for beeing off target (try and error)
t0<-0  # intial simulation time
tmax<-4*3600 # total simulation time (seconds
deltat<-5 # delta t for simulation 

z0<-fit$xm["z0"] # Initial values for simulation 



## Find parameter values for optimal control (before experiment starts)
opt2.null<-nlminb(start = c(p0=0,q0=0,p1=0,q1=0,p2=0,q2=0),
                  objective=obj.fun, Range=Range, t0=t0,
             tmax=tmax,deltat=deltat,
             IntJ0=IntJ0,lambda=lambda, fit = fit,  init.var=init.var,link=link2,z0=fit$xm["z0"],I0=0,
             lower=-10,upper=10,
             control = list(trace = 1,rel.tol=1e-6,x.tol=1e-5))


## Mean value simulation simulation simulation
sim2<-simInt(opt2.null$par,Range,t0,tmax,deltat,lambda,fit,init.var,link2,z0,I0=0)


sim2$IntJ


## Plot strategy
par(mfrow=c(1,2))
matplot(sim2$data$t/3600,cbind(sim2$P),type="l",ylim=Range$P,xlab="Time",ylab="Pressure")
matplot(sim2$data$t/3600,cbind(sim2$Q),type="l",ylim=Range$Q,xlab="Time",ylab="Cross flow")



par(mfrow=c(1,3))
matplot(sim2$data$t/3600,cbind(sim2$sim$state$sim$I)/3600,type="l",xlab="Time",ylab="I")
matplot(sim2$data$t/3600,cbind(sim2$sim$state$sim$z),type="l",xlab="Time",ylab="z")
matplot(sim2$data$t/3600,cbind(sim2$sim$output$sim),type="l",xlab="Time",ylab="flux")



##################################################
# Adaptive control
SimMod<-function(Q,DP,deltat,n,z0,I0,t0){
    ## Simulation
    ## the vectors Q,DP defines the applied crossflow and pressure
    ## deltat define the time increments in Q and DP
    ## n is the niumber of realization
    ## z0 is the initial value of the state (cake layer)
    ## I0 is the integrated flux until time t=t0
    ## t0 is the initial time
    N<-length(Q) # number of datapoints in each realiasation
    Rm<-0.58# Set membrane resistance
    load(file="Model6.Rdata") # load the model
    z<-matrix(ncol=n,nrow=N) # Matrix of hidden state
    J<-matrix(ncol=n,nrow=N) # Matrix of flux
    I<-matrix(ncol=n,nrow=N) # Matrix of integrated flux
    t<-numeric(N)

    ##################################################
    ## Parameter values
    Cb<-Model6$opt$par["Cb"]
    mu<-Model6$opt$par["mu"]
    omegaC<-Model6$opt$par["omegaC"]
    Pa<-Model6$opt$par["Pa"]
    muQ<-Model6$opt$par["muQ"]
    muQ2<-Model6$opt$par["muQ2"]
    sigma<-Model6$opt$par["sigma"]
    sq<-Model6$opt$par["sq"]
    sq2<-Model6$opt$par["sq2"]
    sigP<-Model6$opt$par["sigP"]
    sigQ<-Model6$opt$par["sigQ"]
    PaOmega<-Model6$opt$par["PaOmega"]
    alpha0 <- 0
  ################################
  
  ##################################################
  # Initail values
    z[1, ]<-z0
    J[1, ]<-DP[1]/(Rm + exp(alpha0)*
                   (1+DP[1]/exp(Pa+PaOmega*exp(z[1, ])))*exp(z[1, ]))
  I[1]<-I0
  t[1]<-t0
  ##################################################

  ## Simulation
  for(i in 2:N){
    t[i]<-t[i-1]+deltat
    z[i, ]<-z[i-1, ]+
      (exp(Cb)*exp(-z[i-1])*
       (J[i-1, ]- exp(mu + muQ * Q[i] + muQ2 * Q[i]^2) *
        (1-exp(-exp(omegaC)*exp(z[i-1]))))-
       0.5*exp(2*sigma+2*sigP*DP[i-1]+2*sigQ*Q[i-1]))*deltat+
         exp(sigma+sigP*DP[i-1]+sigQ*Q[i-1])*
           rnorm(n)*sqrt(deltat)
    
    J[i, ]<-DP[i]/(Rm+ exp(alpha0)*(1+DP[i]/exp(Pa+PaOmega*exp(z[i, ])))*exp(z[i, ]))
    I[i, ]<-I[i-1, ]+J[i-1, ]*deltat
  }
  ## return a list with:
  ## Hidden sate z
  ## Flux J
  ## Integrated flux I
  ## Timepoints for the simulation t
  return(list(z=z,J=J,I=I,t=t))
}


##################################################
## Dynamic control
opt2<-opt2.null

sim2<-simInt(opt2.null$par,Range,t0,tmax,deltat,lambda,fit,init.var,link2,z0,I0=0)


## Set seed to be able to reproduce results
set.seed(1223)
z0<-fit$xm["z0"] # set initial value of state
I0<-0# initail value of intgrated flux


## control variables (using 2. order polynomial)
Q<-sim2$data$Q
DP<-sim2$data$DP

## Stochastic similation of process
Sim2<-SimMod(Q,DP,deltat,n=1,z0,I0=0,t0=0)
t<-Sim2$t

## control is a matrix with:
## t: timepoints of simulation
## Q0, DP0: inital control
## Q, DP: updated control
## JM: Mean value simulation of flux
## IM: Mean value simulation of integrated flux
## JS: Stochastic simulation of flux
## IS: Stochastic simulation of integrated flux

control<-cbind(t=t,Q0=Q,DP0=DP,Q=Q,DP=DP,JM0=sim2$sim$output$sim$y,
               IM0=sim2$sim$state$sim[ ,1],
               z0=sim2$sim$state$sim[ ,2],
               JM=sim2$sim$output$sim$y,
               IM=sim2$sim$state$sim[ ,1],
               JS=as.numeric(Sim2$J),
               IS=as.numeric(Sim2$I),sdIM=rev(sim2$sim$state$sd[ ,1])[1],
               zM=sim2$sim$state$sim[ ,2],
               zs=Sim2$z[ ,1])

par(mfrow=c(1,2))
plot(control[ ,"t"],control[ ,"z0"],type="l")
plot(control[ ,"t"],exp(control[ ,"z0"]),type="l")


fit.tmp<-fit

t0<-0;T<-max(t)
t<-seq(t0,T,by=5)

## Matrix for update of parameters
A0<-matrix(c(1,0,-0.5,0,1,0,0,0,1),ncol=3,byrow=TRUE)
A1<-A0

DeltaTC <- 150
imax <- floor(4*3600/DeltaTC)-1
for(i in 1:imax){
  print(imax-i)
  ## Initial values of state variance
   vx0 <- matrix(0, nrow = length(model$states), ncol = length(model$states),
                 dimnames=list(model$states,model$states))
   diag(vx0)<-exp(-10)
  ## diag(vx0)[names(init.var)]<-exp(init.var)
  data<-data.frame(t,y=Sim2$J,DP=DP,Q=Q,VQ2=0,OLQ=0,OLP=0,Rm=0.58)
  ######################################################

  ## Initial state varaibles (updated recursevly)
  fit.tmp$xm[c("z0","I0")]<-c(z0,I0)

  ## Mean values prediction
  pred <- predict(fit.tmp, newdata = data, vx0=vx0,first=TRUE)

  ## New initial time and initial state variables 
  t1<-i*DeltaTC # i*time step (updated every 2.5 minutes)
  i1<-which(abs(t-t1)==min(abs(t-t1)))
  z0<-pred$state$pred[i1 ,2]
  I0<-pred$state$pred[i1 ,1]

  ## update parameters
  Dt<-t1-t0
  t<-seq(t1,T,by=5)
  A1[1,1]<-1;A1[1,2]<-Dt/(T-t1);A1[1,3]<-3/2*Dt^2/(T-t1)^2-0.5
  A1[2,1]<-0;A1[2,2]<-(T-t0)/(T-t1);
  A1[2,3]<-3*Dt*(T-t0)/(T-t1)^2
  A1[3,1]<-0;A1[3,2]<-0;A1[3,3]<-(T-t0)^2/(T-t1)^2
  pars1<-opt2$par
  pars1[c("q0","q1","q2")]<-
    as.numeric(solve(A1)%*%A0%*%opt2$par[c("q0","q1","q2")])
  pars1[c("p0","p1","p2")]<-
    as.numeric(solve(A1)%*%A0%*%opt2$par[c("p0","p1","p2")])
  ##################################################

  ## New optimal parameter for control based on reliasation of process until time t1
  opt2<-nlminb(start = pars1,objective=obj.fun, Range=Range, t0=t1,
               tmax=tmax,deltat=deltat,IntJ0=IntJ0,lambda=lambda,
               fit = fit,init.var=init.var,link=link2,z0=z0,I0=I0,
               lower=-10,upper=10,
               control = list(trace = 1,rel.tol=1e-6,x.tol=1e-5,iter.max=24))
  ## iter.max control the numer of allowed iterrations

  
  ## Mean value simulation 
  sim2<-simInt(opt2$par,Range,t1,tmax,deltat,lambda,fit,init.var,link2,z0,I0)
  ## Stochastic simulation
  Sim2<-SimMod(sim2$Q,sim2$P,deltat,n=1,z0,I0,t0=t1)
  Q<-sim2$data$Q
  DP<-sim2$data$DP
  t<-Sim2$t

  ## Insert values in matix control
  control[control[ ,"t"]>=t1,c("Q","DP","JM","IM","JS","IS","sdIM","zM","zs")]<-
      cbind(sim2$Q,sim2$P,JM=sim2$sim$output$sim$y,
            IM=sim2$sim$state$sim[ ,1],Sim2$J,Sim2$I,sdIM=rev(sim2$sim$state$sd[ ,1])[1],
            zM=sim2$sim$state$sim[ ,2],zs=Sim2$z[ ,1])

  t0<-t1
}

## Plot cake build up
matplot(control[ ,"t"]/3600,exp(control[ ,c("z0","zs")]),type="l")


## Plot control and updated control
par(mfrow=c(1,2))
matplot(control[ ,"t"]/3600,control[ ,c("Q","Q0","DP","DP0")],type="l",
        col=c(1,1,2,2),lwd=c(1,2,1,2),lty=c(1,2),xlab="Time [h]",ylab="Q/P")
matplot(control[ ,"t"]/3600,control[ ,c("JM0","JM","JS")],type="l",ylab="flux",xlab="Time")

##################################################
# End
##################################################

