##################################################
## CTSM-R @: www.ctsm.info
## Simulation of the state
## By Jan Kloppenborg Moller (jkmo@dtu.dk) and Rune Juhl
## Please read the README.txt file of this folder
## License: GPL v2
##################################################


##################################################
# Functions for evaluating the log-likelihood 
##################################################
rm(list=ls())
library(doParallel)
library(parallel)
library(numDeriv)

NLlikelihood <- function(data, pars, fit) {
  # Negative log-likelihood
  
  if (is.list(data) && !is.data.frame(data)) {
    # If data is a list (multiple experiments) the negative
    # log-likelihood is evaluated over all elements
    nll <-
    foreach (i = seq_along(data), .combine=c) %dopar% {
      NLlikelihood(data[[i]], pars, fit)
    }
    nll <- sum(nll)
    return(nll)
  }

  # Huber Psi limit (c), define outliers 
  c <- fit$model$options$hubersPsiLimit
  
  fit$xm[names(pars)] <- pars

  # 1-step predictions using model in fit and first order interpolation.
  pred <- predict(fit, newdata = data, first=TRUE)
  
  y <- data[, fit$model$outputs]# output
  
  ressq <- (y - pred$output$pred[, fit$model$outputs])^2 / pred$output$sd[, fit$model$outputs]^2
  ressq[ressq>c^2]<-c*(2*sqrt(ressq[ressq>c^2])-c) # deal with outliers
  nll<-0.5*sum(log(pred$output$sd[, model$outputs]^2)+ressq)# negative log-likelihood
  return(nll)        
}

# obj.fun and obj.gradient: functions used in optimsation including handling of
#       retruned nll-values equal Inf
obj.fun <- function(pars, data, fit) { 
  
  tryCatch(NLlikelihood(data=data, pars=pars, fit), 
           error = function(e) {
             if (grepl("Code:", conditionMessage(e))) {
               Inf
             } else {
               stop(e)
             }
           }
  )  
}

obj.gradient <- function(pars, data, fit) {
  
  step.size <- (.Machine$double.eps)^(1/3)
  
  g <- 
    foreach (p = pars, i = icount(), .combine = c) %:%
    foreach (d = c(1,-1), .combine = "-") %dopar% {
      obj.fun(replace(pars, i, p + d*step.size), data, fit)
    }
  
  g <- g / (2*step.size)
  names(g) <- names(par)
  g
}

##################################################
# end likelihood definitions
##################################################

##################################################
## Model 1
##################################################
library(ctsmr)
model1 <- ctsm$new()

# System equation
eqs <- list()
eqs[["J"]] <- quote(DP / (Rm + exp(alpha0) * (1 + DP / exp(Pa)) * exp(z)))

model1$alg.eqs <- eqs
model1$addSystem(dz ~ (exp(Cb) * exp(-z) * (J - exp(mu) * (1 - exp(-exp(omegaC) * exp(z)))) -
                         0.5 * exp(2 * sigma + 2 * sigP * DP + 2 * sigQ * Q)) * dt +
                        exp(sigma + sigP * DP + sigQ * Q) * dw4)


# Observation equation
model1$addObs(y ~ J)

# Observation variance
model1$setVariance(y ~ exp(2 * s0) + VQ2 * exp(2 * sq2) + 5 * 10^(-5) * exp(-log(2) / (10 * 60) * t) +
                         OLQ * exp(2 * sq) + OLP * exp(2 * sp))

# Define input variables
model1$addInput("Q")
model1$addInput("DP")
model1$addInput("Rm")
model1$addInput("VQ2")
model1$addInput("OLQ")
model1$addInput("OLP")



ctsmr:::Compile(model1)


fake.fit <- structure(list(xm = structure(rep(0, length(model1$pars)), names = model1$pars),
                           model = model1),
                      class = "ctsmr")

fake.fit$xm["z0"] <- -0.6
fake.fit$xm["Cb"] <- -6.3
fake.fit$xm["mu"] <- -0.8
fake.fit$xm["omegaC"] <- 3.6
fake.fit$xm["Pa"] <- 2.2
fake.fit$xm["sigma"] <- -4.4
fake.fit$xm["sigP"] <- 0.1
fake.fit$xm["sigQ"] <- -0.2
fake.fit$xm["sq"] <- -0.7
fake.fit$xm["sq2"] <- 0.07

fake.fit$xm["s0"] <- -40
fake.fit$xm["sp"] <- -40


model1$options$initialVarianceScaling<-exp(-10)

pars <- fake.fit$xm[c("z0","Cb","mu","omegaC","Pa","sigma","sigP","sigQ","sq","sq2")]

model <- model1
OPT <- nlminb(start = pars, objective=obj.fun, data = data,
              fit = fake.fit, control = list(trace = 1))

## AIC
2 * OPT$objective + 2 * length(OPT$par)



# Save the result
Model1 <- list(opt=OPT,obj.fun,NLlikelihood,model=model1,fit=fake.fit)


H <- hessian(obj.fun, Model1$opt$par, data = data, fit = fake.fit)

colnames(H) <- rownames(H) <- names(Model1$opt$par)
Model1 <- list(opt=OPT, obj.fun, NLlikelihood,model=model1,fit=fake.fit,H=H)

save(Model1,file="Model1.Rdata")

##################################################
## Model 2
##################################################

## Define model struture
library(ctsmr)
model2 <- ctsm$new()

# System equation
eqs <- list()
eqs[["J"]] <- quote(DP / (Rm + exp(alpha0) * (1 + DP / exp(Pa)) * exp(z)))
model2$alg.eqs <- eqs

model2$addSystem(dmu ~ exp(sigmu) * dw3)

model2$addSystem(dz~(exp(Cb) * exp(-z) * (J - exp(mu + muQ * Q + muQ2 * Q^2 + muQ3 * Q^3 +
                                                 muP * DP + muP2 * DP^2 + muP3 * DP^3) *
                                         (1 - exp(-exp(omegaC) * exp(z)))) -
                    0.5 * exp(2 * sigma)) * dt +
                   exp(sigma + sigP * DP + sigQ * Q) * dw4)


## Observation equation
model2$addObs(y ~ J)

## Observation variance
model2$setVariance(y ~ exp(2 * s0) + VQ2 * exp(2 * sq2) + 5 * 10^(-5) * exp(-log(2)/(10*60)*t) +
                      OLQ * exp(2 * sq) + OLP * exp(2 * sp))

## Define input variables
model2$addInput("Q") 
model2$addInput("DP")
model2$addInput("Rm") ## estimated for each membrane

## In put related to unceartainty
model2$addInput("VQ2")
model2$addInput("OLQ")
model2$addInput("OLP")

## compile model
ctsmr:::Compile(model2)

##################################################


## Ctsm setting
model2$options$initialVarianceScaling<-exp(-10)


# Set op an object (fake_fit) which have the same structure as an CTSM-fit,
# Used in optimisation
fake.fit <- structure(list(xm = structure(rep(0, length(model2$pars)), names = model2$pars),
                           model = model2),
                      class = "ctsmr")

## Initial values
fake.fit$xm["z0"] <- -0.6
fake.fit$xm["mu0"] <- -3.5
fake.fit$xm["muQ"] <- 2.2
fake.fit$xm["muQ2"] <- -0.4
fake.fit$xm["Cb"] <- -5.9
fake.fit$xm["omegaC"] <- 3.6
fake.fit$xm["Pa"] <- 2.2
fake.fit$xm["sigma"] <- -4.4
fake.fit$xm["sq"] <- -0.7
fake.fit$xm["sq2"] <- 0.07
fake.fit$xm["sigP"] <- 0.09
fake.fit$xm["sigQ"] <- -0.17
## 

## parameters that is set to zero (i.e. log(par) app zero)
fake.fit$xm["s0"] <- -40
fake.fit$xm["sigmu"] <- -40
fake.fit$xm["sp"] <- -40

library(doParallel)
registerDoParallel(25)# number of cores


## define parameter that are optimised 
pars <- fake.fit$xm[c("mu0", "z0", "Cb", "muQ", "muQ2", "omegaC",
                      "Pa", "sigma", "sigP", "sigQ", "sq", "sq2")]

model <- model2
OPT <- nlminb(start = pars, objective = obj.fun, data = data,
              fit = fake.fit, control = list(trace = 1))

## AIC
OPT$objective * 2 + 2 * length(OPT$par)

# Save the result
Model2 <- list(opt=OPT,obj.fun,NLlikelihood,model=model2,fit=fake.fit)


## Find Hessian (used for evaluating parameter unceartainty)
library(numDeriv)

H <- hessian(obj.fun, Model2$opt$par, data = data, fit = fake.fit)
colnames(H) <- rownames(H) <- names(Model2$opt$par)

## Collect functions and fit
Model2 <- list(opt = Model2$opt, obj.fun = obj.fun, NLlikelihood = NLlikelihood,
                 model = model2, fit = fake.fit, H = H)
save(Model2,file="Model2.Rdata")

##################################################
## Model 3
library(ctsmr)
model3 <- ctsm$new()

# System equation
eqs <- list()
eqs[["J"]] <- quote(DP / (Rm + exp(alpha0) * (1 + DP / exp(Pa)) * exp(z)))

model3$alg.eqs <- eqs


model3$addSystem(dmu ~ exp(sigmu) * dw3)

model3$addSystem(dz ~ (exp(Cb) * exp(-z) * (J - mu/(1 + exp(-alphamu * (Q - Q0)))*
                                     (1 - exp(-exp(omegaC) * exp(z))))-
                    0.5 * exp(2 * sigma + 2*sigP * DP + 2 * sigQ*Q)) * dt +
                exp(sigma + sigP * DP + sigQ * Q) * dw4)


# Observation equation
model3$addObs(y~J)

# Observation variance
model3$setVariance(y ~ exp(2 * s0) + VQ2 * exp(2 * sq2) +
                          5 * 10^(-5) * exp(-log(2) / (10 * 60) * t) +
                          OLQ * exp(2 * sq) + OLP * exp(2 * sp))

# Define input variables
model3$addInput("Q")
model3$addInput("DP")
model3$addInput("Rm")

model3$addInput("VQ2")
model3$addInput("OLQ")
model3$addInput("OLP")

## Compile model
ctsmr:::Compile(model3)


fake.fit <- structure(list(xm = structure(rep(0, length(model3$pars)), names = model3$pars),
                           model = model3), class = "ctsmr")


## Model setting
model3$options$initialVarianceScaling<-exp(-10)

## initial values
fake.fit$xm["mu0"] <- 0.5
fake.fit$xm["z0"] <- -0.6
fake.fit$xm["Cb"] <- -5.8
fake.fit$xm["omegaC"] <- 3.6
fake.fit$xm["Pa"] <- 2.2
fake.fit$xm["alphamu"] <- 6.5
fake.fit$xm["Q0"] <- 1

## initial values related to diffussion and obs variance
fake.fit$xm["sigma"] <- -4.4
fake.fit$xm["sq"] <- -0.6
fake.fit$xm["sq2"] <- -0.07
fake.fit$xm["sigP"] <- 0.1
fake.fit$xm["sigQ"] <- -0.17

fake.fit$xm["s0"] <- -40
fake.fit$xm["sigmu"] <- -40
fake.fit$xm["sp"] <- -40

pars <- fake.fit$xm[c("mu0", "z0", "Cb", "omegaC", "Pa", "alphamu", "Q0",
                   "sigma", "sq", "sq2", "sigP", "sigQ")]


model <- model3
OPT <- nlminb(start = pars, objective=obj.fun, data = data,
              fit = fake.fit, control = list(trace = 1))
## AIC
2 * OPT$objective + 2 * length(OPT$par)

# Save the result
Model3 <- list(opt=OPT, obj.fun, NLlikelihood,model=model3,fit=fake.fit)


## Find Hessian
H <- hessian(obj.fun, Model3$opt$par, data = data, fit = fake.fit)

colnames(H)<-rownames(H)<-names(Model3$opt$par)
Model3 <- list(opt=OPT, obj.fun, NLlikelihood,model=model3,fit=fake.fit,H=H)
save(Model3,file="Model3.Rdata")



##################################################
## Model 4
##################################################

library(ctsmr)
model4 <- ctsm$new()

# System equation
eqs <- list()
eqs[["J"]] <- quote(DP / (Rm + exp(alpha0) * (1 + DP / exp(Pa)) * exp(z)))

model4$alg.eqs <- eqs


model4$addSystem(dmu ~ exp(sigmu) * dw3)


model4$addSystem(dz ~ (exp(Cb) * exp(-z) * (J - mu * Q^gamma *
                                                   (1 - exp(-exp(omegaC) * exp(z))))-
                              0.5 * exp(2 * sigma + 2 * sigP * DP + 2 * sigQ * Q)) * dt +
                             exp(sigma + sigP * DP + sigQ * Q) * dw4)


# Observation equation
model4$addObs(y ~ J)

# Observation variance
model4$setVariance(y ~ exp(2 * s0) + VQ2 * exp(2 * sq2) + 5 * 10^(-5) *
                              exp(-log(2) / (10 * 60) * t) +
                              OLQ * exp(2 * sq) + OLP * exp(2 * sp))

# Define input variables
model4$addInput("Q")
model4$addInput("DP")
model4$addInput("Rm")

model4$addInput("VQ2")
model4$addInput("OLQ")
model4$addInput("OLP")


ctsmr:::Compile(model4)


fake.fit <- structure(list(xm = structure(rep(0, length(model4$pars)),
                                          names = model4$pars),
                           model = model4),
                      class = "ctsmr")

fake.fit$xm["mu0"] <- 0.14
fake.fit$xm["z0"] <- -0.16
fake.fit$xm["Cb"] <- -1.6
fake.fit$xm["gamma"] <- 0
fake.fit$xm["omegaC"] <- 1
fake.fit$xm["Pa"] <- 0.6

fake.fit$xm["sigma"] <- -1.2
fake.fit$xm["sigP"] <- 0
fake.fit$xm["sigQ"] <- 0
fake.fit$xm["sq"] <- -0.2
fake.fit$xm["sq2"] <- 0

fake.fit$xm["sigmu"] <- -40
fake.fit$xm["s0"] <- -40
fake.fit$xm["sp"] <- -40

pars <- fake.fit$xm[c("mu0", "z0", "Cb", "gamma", "omegaC", "Pa", "sigma", "sigP",
                      "sigQ", "sq", "sq2")]

model4$options$initialVarianceScaling<-exp(-10)

model <- model4
OPT <- nlminb(start = pars, objective = obj.fun, data = data,
              fit = fake.fit, control = list(trace = 1))
## AIC
OPT$objective * 2 + 2 * length(OPT$par)

# Save the result
Model4 <- list(opt=OPT, obj.fun, NLlikelihood,model=model4,fit=fake.fit)

H <- hessian(obj.fun, Model4$opt$par, data = data, fit = fake.fit)

colnames(H)<-rownames(H)<-names(Model4$opt$par)
Model4 <- list(opt=OPT, obj.fun, NLlikelihood,model=model4,fit=fake.fit,H=H)

save(Model4,file="Model4.Rdata")


##################################################
## Model 5
##################################################



library(ctsmr)
model5 <- ctsm$new()

# System equation
eqs <- list()
eqs[["J"]] <- quote(DP / (Rm + exp(alpha0) * (1 + DP / exp(Pa + PaI * I/3600 + PaOmega * exp(z))) * exp(z)))
model5$alg.eqs <- eqs

model5$addSystem(dI ~ J * dt + exp(sigI) * dw1)
model5$addSystem(dz ~ (exp(Cb) * exp(-z) * (J - exp(mu + muQ*Q + muQ2 * Q^2 + muQ3 * Q^3 +
                                                      muP * DP + muP2 * DP^2 + muP3 * DP^3) *
                                              (1 - exp(-exp(omegaC) * exp(z))))-
                         0.5 * exp(2 * sigma + 2 * sigP * DP + 2 * sigQ * Q)) * dt +
                      exp(sigma + sigP * DP + sigQ * Q) * dw2)


# Observation equation
model5$addObs(y ~ J)

# Observation variance
model5$setVariance(y ~ exp(2 * s0) + VQ2 * exp(2 * sq2) + 5 * 10^(-5) * exp(-log(2) / (10 * 60) * t)+
                         OLQ * exp(2 * sq) + OLP * exp(2 * sp))

# Define input variables
model5$addInput("Q")
model5$addInput("DP")
model5$addInput("Rm")

model5$addInput("VQ2")
model5$addInput("OLQ")
model5$addInput("OLP")

ctsmr:::Compile(model5)

fake.fit <- structure(list(xm = structure(rep(0, length(model5$pars)), names = model5$pars),
                           model = model5),
                      class = "ctsmr")

model5$options$initialVarianceScaling<-exp(-10)


pars*0.9
fake.fit$xm["z0"] <- -0.5
fake.fit$xm["Cb"] <- -5.3
fake.fit$xm["mu"] <- -3.2
fake.fit$xm["omegaC"] <- 3.2
fake.fit$xm["Pa"] <- 1.5
fake.fit$xm["PaI"] <- 0.5
fake.fit$xm["muQ"] <- 2
fake.fit$xm["muQ2"] <- -0.36

fake.fit$xm["sigma"] <- -4
fake.fit$xm["sigP"] <- 0.1
fake.fit$xm["sigQ"] <- -0.2
fake.fit$xm["sq"] <- -0.6
fake.fit$xm["sq2"] <- 0.06

fake.fit$xm["s0"] <- -40
fake.fit$xm["sp"] <- -40
fake.fit$xm["sigI"] <- -40

pars <- fake.fit$xm[c("z0","Cb","mu","omegaC","Pa","PaI","muQ","muQ2","sigma",
                      "sigP","sigQ","sq","sq2")]

model <- model5
OPT <- nlminb(start = pars, objective=obj.fun, data = data,
              fit = fake.fit, control = list(trace = 1))

## AIC
OPT$objective * 2 + 2 * length(OPT$par)

# Save the result
Model5 <- list(opt=OPT,obj.fun,NLlikelihood,model=model5,fit=fake.fit)

## Hessian
H <- hessian(obj.fun, Model5$opt$par, data = data, fit = fake.fit)
colnames(H) <- rownames(H) <- names(Model5$opt$par)

Model5 <- list(opt=OPT, obj.fun, NLlikelihood,model=model5,fit=fake.fit,H=H)

save(Model5,file="Model5.Rdata")



##################################################
## Model 6
##################################################



library(ctsmr)
model6 <- ctsm$new()

# System equation
eqs <- list()
eqs[["J"]] <- quote(DP / (Rm + exp(alpha0) * (1 + DP / exp(Pa + PaI * I / 3600 +
                                                           PaOmega * exp(z))) * exp(z)))
model6$alg.eqs <- eqs

model6$addSystem(dI ~ J*dt + exp(sigI) * dw1)
model6$addSystem(dz ~ (exp(Cb) * exp(-z) * (J - exp(mu + muQ * Q + muQ2 * Q^2 + muQ3 * Q^3+
                                                      muP * DP + muP2 * DP^2 + muP3 * DP^3) *
                                              (1 - exp(-exp(omegaC) * exp(z)))) -
                         0.5 * exp(2 * sigma + 2 * sigP * DP + 2 * sigQ * Q)) * dt +
                        exp(sigma + sigP * DP + sigQ * Q) * dw2)

# Observation equation
model6$addObs(y ~ J)

# Observation variance
model6$setVariance(y ~ exp(2 * s0) + VQ2 * exp(2 * sq2) + 5 * 10^(-5) *
                         exp(-log(2) / (10 * 60) * t) +
                         OLQ * exp(2 * sq) + OLP * exp(2 * sp))

# Define input variables
model6$addInput("Q")
model6$addInput("DP")
model6$addInput("Rm")

model6$addInput("VQ2")
model6$addInput("OLQ")
model6$addInput("OLP")

ctsmr:::Compile(model6)

fake.fit <- structure(list(xm = structure(rep(0, length(model6$pars)), names = model6$pars),
                           model = model6), class = "ctsmr")

fake.fit$xm["mu"] <- -3.5
fake.fit$xm["z0"] <- -0.7
fake.fit$xm["Cb"] <- -5.9
fake.fit$xm["omegaC"] <- 3.6
fake.fit$xm["Pa"] <- 1.1
fake.fit$xm["muQ"] <- 2.2
fake.fit$xm["muQ2"] <- -0.4
fake.fit$xm["PaOmega"] <- 0.58

fake.fit$xm["sigma"] <- -4.4
fake.fit$xm["sq"] <- -0.7
fake.fit$xm["sq2"] <- 0.07
fake.fit$xm["sigP"] <- 0.2
fake.fit$xm["sigQ"] <- -0.2

fake.fit$xm["s0"] <- -40
fake.fit$xm["sigI"] <- -40
fake.fit$xm["sp"] <- -40

names(pars)
pars <- fake.fit$xm[c("mu", "z0", "Cb", "omegaC", "Pa", "muQ", "muQ2", "sigma",
                      "sq", "sq2", "sigP", "sigQ", "PaOmega")]


model6$options$initialVarianceScaling<-exp(-10)

model <- model6
OPT <- nlminb(start = pars, objective=obj.fun, data = data,
              fit = fake.fit, control = list(trace = 1))
## AIC
2 * OPT$objective + 2 * length(OPT$par)


# Save the result
Model6 <- list(opt=OPT,obj.fun,NLlikelihood,model=model6,fit=fake.fit)

H <- hessian(obj.fun, Model6$opt$par, data = data, fit = fake.fit)

colnames(H) <- rownames(H) <- names(Model6$opt$par)
Model6 <- list(opt=OPT, obj.fun, NLlikelihood,model=model6,fit=fake.fit,H=H)

save(Model6,file="Model6.Rdata")


##################################################
## End
##################################################
