##################################################
## CTSM-R @: www.ctsm.info
## Simulation of the state
## By Jan Kloppenborg Moller (jkmo@dtu.dk)
## Please read the README.txt file
## License: GPL v2
##################################################

set.seed(129348)
x0 <- 1
N <- 40000
T <- 1
t <- seq(0, T, length=N + 1)
dt <- t[2] - t[1]
x <- numeric(length(t))
x[1] <- x0
sigma <- 0.2
mu <- 2
A <- 5
for(i in 1:N){
    x[i+1] <- x[i] + A * (mu - x[i]) * dt + sigma * sqrt(dt) * rnorm(1)
}

plot(t,x,type="l")
### Simulation of the observable
nobs <- 1000
y <- 1 / (1 + x[seq(2, N + 1, by = nobs)])
y <- y + rnorm(length(y), sd = 0.001)
tobs <- t[seq(2, N + 1, by = nobs)]
### Fits by CTSM-R
dat <- data.frame(t = tobs, y = y)
library(ctsmr)
model <- ctsm$new()

## Specify the state equation
model$addSystem(dx ~ A * (mu - x) * dt + exp(sigma) * dw)

# The observation equation
model$addObs( y ~ 1 / (1 + x) )
model$setVariance(y ~ exp(s))

## Parameter settings
model$setParameter( x0 = c(init = 1, lb = 0, ub = 2))
model$setParameter( A = c(init = 0.1, lb = 0, ub = 10))
model$setParameter( mu = c(init = 1, lb = 0 , ub = 4))
model$setParameter( sigma = c(init = -1.6, lb = -10, ub = 1))
model$setParameter( s = c(init = -8, lb = -20 , ub = 1))

## Fit model m1
fit1 <- model$estimate(data=dat)

## Fit m2 
model$setParameter( mu = c(init = 1.8))
fit2 <- model$estimate(data=dat)

## Fit m3 (ODE)
model$setParameter (sigma = c(init = -10))
model$setParameter (mu = c(init = 1, lb = 0, ub = 4))
fit3 <- model$estimate(data=dat)

## One-step ahead m1
pred1 <- predict(fit1)[[1]]

## Long-term m1
sim1 <- simulate(fit1)[[1]]

## One-step and long term m2
pred2 <- predict(fit2)[[1]]
sim2 <- simulate(fit2)[[1]]

## One-step and long term m3
pred3 <- predict(fit3)[[1]]
sim3 <- simulate(fit3)[[1]]

### Graphs
## panel a)
par(mfrow = c(2, 2),
mar=c(4, 4, 0, 0), oma = c(0, 0, 4, 4))
plot(tobs, y, pch = 19, cex.lab = 1.5, axes = FALSE, xlab="",
     ylab = expression(italic("y")))
polygon(c(tobs, rev(tobs)), c(sim1$output$sim$y +
                              2 * sim1$output$sd$y,
                              rev(sim1$output$sim$y -
                                  2 * sim1$output$sd$y)),
        border=FALSE,col=gray(0.75))
points(tobs, y, pch = 19, cex = 1, col = 1)
box(lwd = 3)
axis(2, lwd = 3, cex.axis = 1.5)

## panel b)
plot(tobs, y, pch = 19, cex = 1.5, col = "white", xlab = "",
     axes = FALSE, ylab = "")
polygon(c(tobs, rev(tobs)), c(sim1$output$sim$y +
                              2 * sim1$output$sd$y,
                              rev(sim1$output$sim$y -
                                  2*sim1$output$sd$y)),
        border = FALSE, col = gray(0.75))
points(tobs, pred1$output$pred$y, col = 2, pch = 19)
points(tobs, pred2$output$pred$y, col = 4, pch = 19)
lines(tobs, sim1$output$sim$y, col = 2, lwd = 2, lty = 1)
lines(tobs, sim2$output$sim$y, col = 4, lwd = 3, lty = 1)
lines(tobs, sim3$output$sim$y, col = 'yellow', lwd = 3,lty = 1)
box(lwd = 3)
axis(4, lwd = 3, cex.axis = 1.5)

## panel c)
plot(t, x, type = "l", axes = F, xlab = expression(italic("t")),
     col="white", ylab = expression(italic("x")), cex.lab = 1.5)
polygon(c(tobs, rev(tobs)), c(sim1$state$sim$x + 2 * sim1$state$sd$x,
                              rev(sim1$state$sim$x - 2 * sim1$state$sd$x)),
        border = FALSE, col = gray(0.75))
lines(t, x, type = "l", col = "red")
lines(tobs, sim3$state$sim$x, col = 'yellow', lwd = 3, lty = 1)
box(lwd = 3)
axis(2, lwd=3, cex.axis=1.5)
axis(1, lwd=3, cex.axis=1.5)

## panel d)
plot(t, x, type = "l", axes = F, xlab = expression(italic("t")),
     col = "white", ylab = "", cex.lab = 1.5)
polygon(c(tobs, rev(tobs)), c(sim1$state$sim$x + 2 * sim1$state$sd$x,
                              rev(sim1$state$sim$x - 2 * sim1$state$sd$x)),
        border = FALSE, col = gray(0.75))
lines(tobs, sim1$state$sim$x, col = "red", lwd = 2, lty = 1)
lines(tobs, pred1$state$pred$x, col = "red", pch = 19)
lines(tobs, pred2$state$pred$x, col = 4, pch = 19)
lines(tobs, sim2$state$sim$x, col = 4, lwd = 3, lty = 1)
lines(tobs, pred3$state$pred$x, col = 'yellow', pch = 19)
lines(tobs, sim3$state$sim$x, col = 'yellow', lwd = 3, lty = 1)
box(lwd = 3)
axis(4, lwd = 3, cex.axis = 1.5)
axis(1, lwd = 3, cex.axis = 1.5)

### Log likelihood and RMSE

## fit1
fit1$loglik
RMSE = function(mod, obs){
  sqrt(mean((mod - obs)^2))
}
RMSE(sim1$output$sim$y, dat$y)

### parameters
## fit1

summary(fit1, extended=TRUE)

## confidence intervals (2 st.d.)
fit1$xm['x0'] + c(-1.96, 1.96) * fit1$sd['x0']

## Sigma and s
exp(fit1$xm['sigma'] + c(-1.96, 1.96) * fit1$sd['sigma'])
sqrt(exp(fit1$xm['s'] + c(-1.96,1.96) * fit1$sd['s']))
# similarly for fit2 and fit3
